"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from json import dumps
from json import loads
from typing import Any
from typing import AsyncIterator
from typing import Iterator

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text

from httpx import AsyncByteStream
from httpx import Response
from httpx import SyncByteStream

from pytest import fixture
from pytest import mark

from respx import MockRouter

from . import SAMPLES
from ..bridge import Bridge
from ..params import BridgeParams



@fixture
def bridge() -> Bridge:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = BridgeParams(
        server='192.168.1.10',
        token='mocked')

    return Bridge(params)



def test_Bridge(
    bridge: Bridge,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    """


    attrs = list(bridge.__dict__)

    assert attrs == [
        '_Bridge__params',
        '_Bridge__client']


    assert inrepr(
        'bridge.Bridge object',
        bridge)

    assert hash(bridge) > 0

    assert instr(
        'bridge.Bridge object',
        bridge)


    assert bridge.params is not None

    assert bridge.client is not None



def test_Bridge_request(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _source = read_text(
        f'{SAMPLES}/source'
        '/resource.json')

    location = (
        'https://192.168.1.10')


    (respx_mock
     .get(f'{location}/clip/v2/resource')
     .mock(Response(
         status_code=200,
         content=_source)))


    response = (
        bridge.request(
            'get', 'resource'))

    response.raise_for_status()


    fetched = response.json()

    sample_path = (
        f'{SAMPLES}/dumped'
        '/resource.json')

    sample = load_sample(
        sample_path, fetched,
        update=ENPYRWS)

    expect = prep_sample(fetched)

    assert sample == expect



def test_Bridge_events_block(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _events = loads(read_text(
        f'{SAMPLES}/source'
        '/events.json'))

    location = (
        'https://192.168.1.10'
        '/eventstream/clip/v2')


    class ByteStream(SyncByteStream):

        def __iter__(
            self,
        ) -> Iterator[bytes]:

            chunks = [
                (f'data: {dumps(x)}\n'
                 .encode('utf-8'))
                for x in _events]

            chunks.insert(0, b': hi\n')

            yield from chunks


    streamer = ByteStream()

    (respx_mock
     .get(location)
     .mock(Response(
         status_code=200,
         stream=streamer)))


    events = list(
        bridge.events_block())

    chunks: list[dict[str, Any]] = []

    for chunk in events:
        chunks.extend(chunk)


    sample_path = (
        f'{SAMPLES}/dumped'
        '/events.json')

    sample = load_sample(
        sample_path, chunks,
        update=ENPYRWS)

    expect = prep_sample(chunks)

    assert sample == expect



@mark.asyncio
async def test_Bridge_events_async(
    bridge: Bridge,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param social: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _events = loads(read_text(
        f'{SAMPLES}/source'
        '/events.json'))

    location = (
        'https://192.168.1.10'
        '/eventstream/clip/v2')


    class ByteStream(AsyncByteStream):

        async def __aiter__(
            self,
        ) -> AsyncIterator[bytes]:

            chunks = [
                (f'data: {dumps(x)}\n'
                 .encode('utf-8'))
                for x in _events]

            chunks.insert(0, b': hi\n')

            await asyncio.sleep(0)

            for chunk in chunks:

                yield chunk

                await asyncio.sleep(0)

            await asyncio.sleep(0)


    streamer = ByteStream()

    (respx_mock
     .get(location)
     .mock(Response(
         status_code=200,
         stream=streamer)))


    events = (
        bridge.events_async())

    chunks: list[dict[str, Any]] = []

    async for chunk in events:
        chunks.extend(chunk)


    sample_path = (
        f'{SAMPLES}/dumped'
        '/events.json')

    sample = load_sample(
        sample_path, chunks,
        update=ENPYRWS)

    expect = prep_sample(chunks)

    assert sample == expect



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional

from pydantic import BaseModel



class RedditParams(BaseModel, extra='forbid'):
    """
    Process and validate the class configuration parameters.

    :param timeout: Timeout when waiting for server response.
    :param username: Username for authenticating with server.
    :param password: Password for authenticating with server.
    :param client: Token used when authenticating to server.
    :param secret: Token used when authenticating to server.
    :param useragent: Reddit has specifics about this value.
    :param ssl_verify: Require valid certificate from server.
    :param ssl_capem: Optional path to certificate authority.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    timeout: int = 30

    username: str
    password: str
    client: str
    secret: str
    useragent: str

    ssl_verify: bool = True
    ssl_capem: Optional[str] = None



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text

from httpx import Response

from pytest import fixture

from respx import MockRouter

from . import SAMPLES
from ..params import RouterParams
from ..router import Router



@fixture
def router() -> Router:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    params = RouterParams(
        server='192.168.1.1',
        username='mocked',
        password='mocked')

    return Router(params)



def test_Router(
    router: Router,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param router: Class instance for connecting to service.
    """


    attrs = list(router.__dict__)

    assert attrs == [
        '_Router__params',
        '_Router__client']


    assert inrepr(
        'router.Router object',
        router)

    assert hash(router) > 0

    assert instr(
        'router.Router object',
        router)


    assert router.params is not None

    assert router.client is not None



def test_Router_request(
    router: Router,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param router: Class instance for connecting to service.
    :param respx_mock: Object for mocking request operation.
    """


    _source = read_text(
        f'{SAMPLES}/source.json')

    location = (
        'https://192.168.1.1')


    (respx_mock
     .get(
         f'{location}/proxy/network'
         '/api/s/default/rest/user')
     .mock(side_effect=[
         Response(
             status_code=401),
         Response(
             status_code=200,
             content=_source)]))

    (respx_mock
     .post(f'{location}/api/auth/login')
     .mock(Response(
         status_code=200,
         content=_source)))


    response = (
        router.request_proxy(
            'get', 'rest/user'))

    response.raise_for_status()


    fetched = response.json()

    sample_path = (
        f'{SAMPLES}/dumped.json')

    sample = load_sample(
        sample_path, fetched,
        update=ENPYRWS)

    expect = prep_sample(fetched)

    assert sample == expect



"""
Functions and routines associated with Enasis Network Remote Connect.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from json import loads
from threading import Event as BlockEvent
from typing import Any
from typing import AsyncIterator
from typing import Iterator
from typing import Optional
from typing import TYPE_CHECKING

from httpx import Response

from ..utils import HTTPClient
from ..utils.http import _METHODS

if TYPE_CHECKING:
    from .params import BridgeParams

AsyncEvent = asyncio.Event



class Bridge:
    """
    Interact with the cloud service API with various methods.

    :param params: Parameters for instantiating the instance.
    """

    __params: 'BridgeParams'
    __client: HTTPClient


    def __init__(
        self,
        params: 'BridgeParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__params = params

        client = HTTPClient(
            timeout=params.timeout,
            verify=params.ssl_verify,
            capem=params.ssl_capem)

        self.__client = client


    @property
    def params(
        self,
    ) -> 'BridgeParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def client(
        self,
    ) -> HTTPClient:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def request(
        self,
        method: _METHODS,
        path: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> Response:
        """
        Return the response for upstream request to the server.

        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        :returns: Response for upstream request to the server.
        """

        params = dict(params or {})
        json = dict(json or {})

        server = self.params.server
        token = self.params.token
        client = self.client

        token_key = 'hue-application-key'

        location = (
            f'https://{server}'
            f'/clip/v2/{path}')

        request = client.request_block

        return request(
            method=method,
            location=location,
            params=params,
            headers={token_key: token},
            json=json)


    def events_block(
        self,
        timeout: int = 60,
        cancel: Optional[BlockEvent] = None,
    ) -> Iterator[list[dict[str, Any]]]:
        """
        Return the response for upstream request to the server.

        :param timeout: Timeout when waiting for server response.
        :param cancel: Event used to indicate we should cacnel.
        """

        if cancel is None:
            cancel = BlockEvent()

        server = self.params.server
        token = self.params.token
        client = self.client

        accept = 'text/event-stream'
        token_key = 'hue-application-key'

        location = (
            f'https://{server}'
            f'/eventstream/clip/v2')

        request = client.stream_block


        stream = request(
            method='get',
            location=location,
            headers={
                'Accept': accept,
                token_key: token},
            timeout=timeout)


        for event in stream:

            if event is None:
                continue  # NOCVR

            event = event.strip()

            if event[0:5] != 'data:':
                continue

            yield loads(event[6:])

            if cancel.is_set():
                break  # NOCVR


    async def events_async(
        self,
        timeout: int = 60,
        cancel: Optional[AsyncEvent] = None,
    ) -> AsyncIterator[list[dict[str, Any]]]:
        """
        Return the response for upstream request to the server.

        :param timeout: Timeout when waiting for server response.
        :param cancel: Event used to indicate we should cacnel.
        """

        if cancel is None:
            cancel = AsyncEvent()

        server = self.params.server
        token = self.params.token
        client = self.client

        accept = 'text/event-stream'
        token_key = 'hue-application-key'

        location = (
            f'https://{server}'
            f'/eventstream/clip/v2')

        request = client.stream_async


        stream = request(
            method='get',
            location=location,
            headers={
                'Accept': accept,
                token_key: token},
            timeout=timeout)


        async for event in stream:

            if event is None:
                continue  # NOCVR

            event = event.strip()

            if event[0:5] != 'data:':
                continue

            yield loads(event[6:])

            if cancel.is_set():
                break  # NOCVR


        await asyncio.sleep(0)
